
/*
		cgi/g77 functions library
		-------------------------

		http://krabulator.free.fr/devel/cgi77.html
		mail: oulala@chez.com

	This library is released under the LGPL license.
	Use it as you want, but remember that: "ymmv".

	* only tested under GNU/Linux glibc kernel 2.2.x
	* compilers versions:
	* gcc version egcs-2.91.66 19990314 (egcs-1.1.2 release)
	* g77 version egcs-2.91.66 19990314 (egcs-1.1.2 release)

*/

#include  <stdio.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <ctype.h>

#include  "cgi77.h"

/* ------------------------------------------------------------------- */

static	int	initialized = 0;
static	char	*requete;

typedef struct
	{
	char	*name;
	int	len_name;
	char	*value;
	int	len_value;
	} parameter;

static	parameter	parameters[CGI77_MAX_PARAMETERS];
static	int		nbre_parameters = 0;
static	char		*tmp_par[CGI77_MAX_PARAMETERS];

static  char		*empty_string = "";

/* ------------------------------------------------------------------- */
/*
 *	this function take care of getting the request's parameters
 *	from the adequate location: envvar for GET and stdin for POST
 */

static int
initialization(void)
{
int	methode, taille, numq, foo, bar;
char	*cptr;

methode = cgi77_method__();

#if TRACE
fprintf(stderr, "cgi77: initialization, method %d\n", methode);
#endif


switch (methode)
	{
	case CGI77_METHOD_GET:
		if ((cptr = getenv("QUERY_STRING")) == NULL)
			{
#if TRACE
			fprintf(stderr, "cgi77: init GET: no query string\n");
#endif
			return -1;
			}
		else
			{
			requete = strdup(cptr);
			}
		break;

	case CGI77_METHOD_POST:

		if ((cptr = getenv("CONTENT_LENGTH")) != NULL)
			{
			if (sscanf(cptr, "%d", &taille)==1)
				{
				/*
				 * OK, try to read the query...
				 */
#if TRACE
				fprintf(stderr, "cgi77: POST szquery = %d\n",
								taille);
#endif
				if ((requete=malloc(taille+1)) != NULL)
					{
					foo = read(0, requete, taille);
#if TRACE
					fprintf(stderr, "cgi77: read %d bytes\n", foo);
#endif
					}
				else
					{
					fprintf(stderr, "cgi77: POST no memory\n");
					return -1;
					}
				}
			else
				{
#if TRACE
				fprintf(stderr, "cgi77: bad CONTENT_LENGTH\n");
#endif
				return -1;
				}
			}
		else
			{
#if TRACE
			fprintf(stderr, "cgi77: no CONTENT_LENGTH\n");
#endif
			return -1;
			}
		break;

	default:
		fprintf(stderr, "cgi77: bad method %d in init\n", methode);
		return -1;
		break;
	}

/*
 *	OK, here whe have now the query string in requete[]
 *
 *	and whe have to truncate it to the first <SPACE> ?
 *	Yes I thinks. Bug revelated by w3m, or a bug in w3m ?
 *
 */
cptr = strchr(requete, ' ');
if ( cptr != NULL )
	*cptr = '\0';

#if TRACE
fprintf(stderr, "cgi77: initialization: full request [%s]\n", requete);
#endif

/*
 *	and we now have the m155i0n to parse it ;-)
 */
numq = 0;
cptr = strtok(requete, "&");
while (cptr != NULL)
	{
	tmp_par[numq] = strdup(cptr);
	if (numq++ == CGI77_MAX_PARAMETERS)
		{
		break;
		}
	cptr = strtok(NULL, "&");
	}

nbre_parameters = numq;

#if TRACE
fprintf(stderr, "cgi77: decoded %d parameters\n", nbre_parameters);
#endif

/*
 *	where is the right place in flow for converting
 *	the '+' to ' ' ? just before unescaping ?
 */
for (foo=0; foo<numq; foo++)
	{
	parameters[foo].name = strtok(tmp_par[foo], "=");
	cgi77_unescape(parameters[foo].name);
	if ((parameters[foo].value = strtok(NULL, "=")) != NULL)
		{
		for (bar=0; bar<strlen(parameters[foo].value); bar++)
			{
			if (parameters[foo].value[bar] == '+')
				{
				parameters[foo].value[bar] = ' ';
				}
			}
		cgi77_unescape(parameters[foo].value);
		}
	else
		{
		parameters[foo].value = empty_string;
		/*
		 *	Was it the right thing to do ?
		 */
		}
#if TRACE
	fprintf(stderr, "cgi77: idx %d name [%s] value [%s]\n", 
				foo,
				parameters[foo].name,
				parameters[foo].value);
#endif
	}

initialized = 42;	/*  _always_ use a magic number  */

return 0;
}

/* ------------------------------------------------------------------- */
/*
 *	this function is the generic interface for getting _one_
 *	parameter from the request. you can call it only from C.
 *	
 *	return value:	a pointer on an asciiz string on success,
 *			or the NULL value if not found.
 */
char *
cgi77_get_one_parameter(char *nom, int sznom)
{
int	foo, bar;
char	*buffnom;

if ( ! initialized )
	{
	foo = initialization();
	if (foo)	return 0;
	}
/*
 *	we copy the fortran name of the parameter in an
 *	asciiz temporary buffer.
 */
if ((buffnom = malloc(sznom+1)) == NULL)
	{
	fprintf(stderr, "cgi77: memory error\n");
	return NULL;
	}
for (foo=0; foo<sznom; foo++)
	buffnom[foo] = nom[foo];
buffnom[foo] = '\0';

#if TRACE
fprintf(stderr, "cgi77: param name [%s]\n", buffnom);
#endif

for (bar=0; bar<nbre_parameters; bar++)
	{
	if ( ! strcmp(buffnom, parameters[bar].name) )
		{
#if TRACE
		fprintf(stderr, "cgi77: getparam [%s] found at %d\n",
							buffnom, bar);
#endif
		return parameters[bar].value;
		}
	}

#if TRACE
fprintf(stderr, "cgi77: getparam [%s] NOT found\n", nom);
#endif
return NULL;
}
/* ------------------------------------------------------------------- */

static char
x2c (const char *what)
{
char	digit;


digit  = ((what[0]>='A') ? ((what[0]&0xdf)-'A')+10 : (what[0]-'0'));
digit <<= 4;
digit += ((what[1]>='A') ? ((what[1]&0xdf)-'A')+10 : (what[1]-'0'));

return digit;
}

int
cgi77_unescape(char *str)
{
int	x, y, badesc, badpath;

badesc = badpath = 0;

for (x=y=0; str[y]; ++x, ++y)
	{
	if (str[y] != '%')
		str[x] = str[y];
	else
		{
		if (!isxdigit(str[y+1]) || !isxdigit(str[y+2]))
			{
			badesc = 1;
			str[x] = '%';
			}
		else
			{
			str[x] = x2c(&str[y+1]);
			y += 2;
			if (str[x] == '/' || str[x] == '\0')
				badpath = 1;
			}
		}
	}  /* end of for */

str[x] = '\0';

return badesc+badpath;
}

/* ------------------------------------------------------------------- */

void
cgi77_marker__(char *pstr, int strsize)
{
int	pid, foo;
char	*pbuffer, c;

pid = getpid();

#if TRACE
// fprintf(stderr, "\ncgi77: marker: strsize = %d\n", strsize);
#endif

pbuffer = (char *)malloc(strsize+1);
foo = 0;

while ( foo!=strsize && (c=pstr[foo])!=' ' && c)
	{
#if TRACE
//	fprintf(stderr, "foo %d c %c\n", foo, c);
#endif
	pbuffer[foo++] = c;
	}

#if TRACE
// fprintf(stderr, "cgi77: marker: poking a '\\0' at %d\n", foo);
#endif 

fprintf(stderr, "cgi77: lib v %s - pid %d - %s ---------------------\n",
			CGI77_VERSION, pid, pbuffer);
free(pbuffer);
}
/* ------------------------------------------------------------------- */

